
from abc import abstractmethod
from typing import List, override

from angr import SimState

from dAngr.angr_ext.std_tracker import StdTracker
from dAngr.cli.grammar.definitions import FunctionDefinition
from dAngr.cli.models import Breakpoint
from dAngr.exceptions import ExecutionError


class Filter:
    
    @property
    @abstractmethod
    def enabled(self)->bool:
        pass
    
    @enabled.setter
    @abstractmethod
    def enabled(self, value:bool):
        pass

    def filter(self, state:SimState):
        if self.enabled:
            return self._filter(state)
        return False
    
    @abstractmethod
    def _filter(self, state:SimState)->bool:
        pass


class FilterList(Filter):
    def __init__(self, filters:List[Filter]):
        super().__init__()
        self.filters = filters 
        self._combination = any
    
    @property
    def enabled(self)->bool:
        return all(f.enabled for f in self.filters)
    
    @enabled.setter
    def enabled(self, value:bool):
        for f in self.filters:
            f.enabled = value

    def get_matching_filter(self, state:SimState):
        for f in self.filters:
            if f.filter(state):
                yield f
                
    def append(self, filter:Filter):
        self.filters.append(filter)
    def remove(self, filter:Filter):
        self.filters.remove(filter)
    def pop(self, index:int):
        return self.filters.pop(index)
    def find(self, filter_type:type, condition):
        return next((f for f in self.filters if isinstance(f, filter_type) and condition), None)
    
    def clear(self):
        self.filters.clear()
    def __getitem__(self, index:int):
        return self.filters[index]
    def _filter(self, state:SimState):
        return self._combination(f._filter(state) for f in self.filters)
    def __len__(self):
        return len(self.filters)
    def __repr__(self) -> str:
        return f"FilterList: {', '.join(str(f) for f in self.filters)}"
    def __iter__(self):
        return iter(self.filters)
    def __len__(self):
        return len(self.filters)
    def empty(self):
        return len(self.filters) == 0
class OrFilterList(FilterList):
    def __init__(self, filters:List[Filter]):
        super().__init__(filters)
        self._combination = any
    def __repr__(self) -> str:
        return f"OrFilterList: {', '.join(str(f) for f in self.filters)}"
class AndFilterList(FilterList):
    def __init__(self, filters:List[Filter]):
        super().__init__(filters)
        self._combination = all
    def __repr__(self) -> str:
        return f"AndFilterList: {', '.join(str(f) for f in self.filters)}"
    
class FilterFunction(Filter):
    def __init__(self, func, debugger):
        super().__init__()
        self.func:FunctionDefinition = func
        self.debugger = debugger
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    def _filter(self, state:SimState):
        try:
            prev = self.debugger.current_state
            self.debugger.current_state = state
            return self.func(self.debugger.context)
        finally:
            self.debugger.current_state = prev
    
    def __repr__(self) -> str:
        return f"Filter Function: {self.func.name}"
    
class AddressFilter(Filter):
    def __init__(self, address:int):
        super().__init__()
        self.address = address
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    def _filter(self, state:SimState)->bool:
        if self.address == state.addr:
            return True
        # check if in range of block
        start:int = state.addr # type: ignore
        #check if start is in binary
        if not state.project.loader.main_object.contains_addr(start): # type: ignore
            return False
        instrs = state.block().instruction_addrs
        if not instrs:
            return False
        end = instrs[-1] # type: ignore
        return start <= self.address <= end
    
    def __repr__(self) -> str:
        return f"Address Filter: {hex(self.address)}"

class FunctionFilter(Filter):
    def __init__(self, function_name):
        super().__init__()
        self.function_name = function_name
        self.f_addr = None
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    def _filter(self, state:SimState):
        if self.f_addr is None:
            self.f_addr = func = state.project.kb.functions(name=self.function_name) # type: ignore
        return self.f_addr == state.callstack.func_addr
    
    def __repr__(self) -> str:
        return f"Function Filter: {self.function_name}"
    
class StdStreamFilter(Filter):
    mapping = {0: 'stdin', 1: 'stdout', 2: 'stderr'}
    #stdin, stdout, stderr: 0, 1, 2
    def __init__(self, stream:int, value:str, regex:bool = False):
        super().__init__()

        self.stream = stream
        self.value = value
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    def _filter(self, state:SimState):
        if mapped := self.mapping.get(self.stream):
            std:StdTracker = state.get_plugin(f'{mapped}_tracker')
            std_data = std.get_prev_string()

            return self.value in std_data
        raise ExecutionError(f"Stream {self.stream} not found.")
    
    def __repr__(self) -> str:
        return f"Standard Stream Filter: {self.stream}"
    
class InputFileFilter(Filter):
    def __init__(self, path, value:str):
        super().__init__()
        self.path = path
        self.value = value
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    def _filter(self, state:SimState):
        return self.value in str(state.posix.dump_file_by_path(self.path))
    
    def __repr__(self) -> str:
        return f"Input File Filter: {self.path}"

class SourceFilter(AddressFilter):
    def __init__(self, address:int, source_file:str, line_nr:int):
        super().__init__(address=address)
        self.source_file = source_file
        self.line_nr = line_nr
        
    def __repr__(self) -> str:
        return f"Source Filter: {self.source_file}:{self.line_nr} ({hex(self.address)})"
    
class SymbolicFilter(AddressFilter):
    def __init__(self, address:int):
        super().__init__(address=address)
        self._enabled = True
    
    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    def enabled(self, value:bool):
        self._enabled = value

    @override
    def _filter(self, state:SimState):
        a = state.mem[self.address].int.resolved
        return state.solver.symbolic(a)
    
    def __repr__(self) -> str:
        return f"Symbolic Filter: memory address {hex(self.address)}"




from textwrap import indent
from typing import Any
from claripy import List
from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.cli.grammar.expressions import BASECommand, VariableRef
from dAngr.exceptions import InvalidArgumentError
from dAngr.utils import undefined

class Definition:
    def __init__(self, name):
        self.name = name
    def __call__(self, context:ExecutionContext, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError
    def __repr__(self):
        return f"def {self.name}"
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Definition) and self.name == value.name

class ArgumentSpec:
    def __init__(self, name:str,  dtype:type = type[Any], default=undefined, description:str="") -> None:
        self._name = name
        self._default = default
        self._dtype = dtype
        self._description = description
    @property
    def name(self):
        return self._name
    @property
    def isOptional(self):
        return self._default is not None
    @property
    def default(self):
        return self._default
    @property
    def dtype(self):
        return self._dtype
    @property
    def description(self):
        return self._description
    def __repr__(self):
        r = self.name
        if self.dtype != type[Any]:
            if isinstance(self.dtype, object):
                r += f":{self.dtype.__name__}"
            else:
                r += f":{self.dtype}"
        if self.default != undefined:
            r += f"={self.default}"
        if self.description:
            r += f" {self.description}"
        return r

    def __eq__(self, value: object) -> bool:
        return isinstance(value, ArgumentSpec) and self.name == value.name and self.dtype == value.dtype and self.default == value.default and self.description == value.description

#Definitions    
class FunctionDefinition(Definition):
    def __init__(self, name, package:str|None, args:List[ArgumentSpec]):
        self._name = name
        self._package = package if package else ""
        self._args:List[ArgumentSpec] = args
    @property
    def name(self):
        return self._name
    @property
    def package(self):
        return self._package
    @property
    def args(self):
        return self._args
    @property
    def required_arguments(self) -> List[ArgumentSpec]:
        return [a for a in self._args if a.default == undefined]
    @property
    def optional_arguments(self) -> List[ArgumentSpec]:
        return [a for a in self._args if a.default != undefined]
    
    def get_arg_by_name(self, name:str) -> ArgumentSpec:
        for arg in self._args:
            if arg.name == name:
                return arg
        if n:= next((arg for arg in self._args if arg.name == 'kwargs'), None):
            return arg
        
        raise InvalidArgumentError(f"Unknown argument {name}")
    def __repr__(self):
        return f"def {self.name}({",".join([str(a) for a in self.args])})"
    
    def __eq__(self, value: object) -> bool:
        return isinstance(value, FunctionDefinition) and self.name == value.name and self.args == value.args

#declaration
class Body:
    def __call__(self, context):
        raise NotImplementedError
    
class FunctionContext(ExecutionContext):
    def __init__(self, function:FunctionDefinition, parent=None):
        super().__init__(parent)
        self.function = function
    
class CustomFunctionDefinition(FunctionDefinition):
    def __init__(self, name, args:List[ArgumentSpec], body):
        super().__init__(name, None, args)
        self.body:Body = body


    def __call__(self, context, *arg_values, **named_args):
        # add args
        context = FunctionContext(self, context)
        # match arg_values with required and optional args
        for i, arg in enumerate(arg_values):
            context[self.args[i].name]= arg_values[i]
        for k,v in named_args.items():
            context[k] = v
            
        result = self.body(context)
        if isinstance(result, BASECommand) and result.base == "return":
            result = result.return_value
        # remove args
        for arg in self.args:
            if isinstance(arg, VariableRef):
                del context.variables[arg.name]
        return result

    def __repr__(self):
        f = super().__repr__()
        return f"{f}\n{indent(str(self.body),"   ")}"
    
    def __eq__(self, value: object) -> bool:
        return isinstance(value, CustomFunctionDefinition) and self.name == value.name and self.args == value.args and self.body == value.body


import html
import os

from prompt_toolkit import PromptSession

from prompt_toolkit.completion import WordCompleter, merge_completers, PathCompleter
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit import HTML
from prompt_toolkit.document import Document
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.history import FileHistory

from dAngr.cli import  CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection
from dAngr.cli.command_line_debugger import DEBUGGER_COMMANDS, dAngrExecutionContext

from dAngr.cli.debugger_commands.base import BuiltinFunctionDefinition
from dAngr.cli.script_processor import ScriptProcessor

# add logger
from dAngr.utils.loggers import get_logger
logger = get_logger(__name__)


class MyValidator(Validator):
    def __init__(self, dbg):
        self.dbg = dbg
    def validate(self, document: Document):
        if e := self.dbg.validate_input(document.text):
            if "INDENT" in e or "&newline" in e:
                return
            raise ValidationError(message=e)


DEBUG_COMMANDS = False
class Server:
    def __init__(self, debug_file_path = None, script_path=None):
        logger.info("Initializing dAngr server with debug_file_path: %s and script_path: %s", debug_file_path, script_path)
        self.commands = DEBUGGER_COMMANDS
        self.completer = None
        self.reset_completer()
        self.debug_file_path = debug_file_path
        self.script_path = script_path
        self.stop = False

    def reset_completer(self, context:dAngrExecutionContext|None=None):
        if context:
            #fetch variables, symbols, and functions with both name and shortname from the context
            dd = {v: f"&vars.{v}" for v in context.variables.keys()}
            dd.update({f"&vars.{v}": f"&vars.{v}" for v in context.variables.keys()})
            dd.update({f"&sym.{s}": f"&sym.{s}" for s in context.debugger._symbols.keys()})
            dd.update({s: f"&sym.{s}" for s in context.debugger._symbols.keys()})
            if context.debugger.initialized:
                dd.update({f"&reg.{s}": f"&reg.{s}" for s in context.debugger.list_registers().keys()})
            dd.update({f"{f.name}": f"{f.name}" for f in context.functions.values()})
            if context.debugger.initialized:
                dd.update({f"&state": f"&state"})
            for f in context.functions.values():
                if isinstance(f, BuiltinFunctionDefinition):
                    dd.update({f"{f.short_name}": f"{f.short_name}"})
        else:
            dd = {c: f"{c} ({self.commands[c].short_name})" for c in self.commands.keys()}
            dd.update({self.commands[c].package + '.' + c: f"{c}" for c in self.commands.keys()})

        word_completer = WordCompleter(sorted(dd.keys()), display_dict=dd, WORD=True)
        self.completer = merge_completers([word_completer,PathCompleter(get_paths=lambda: [os.getcwd()])])


    def loop(self):
        conn = CliConnection()
        dbg = CommandLineDebugger(conn)
        conn.send_info("Welcome to dAngr, the symbolic debugger. Type help or ? to list commands.")
        prmpt = HTML('<darkcyan>(dAngr)> </darkcyan>')
        cmd_history = FileHistory(os.path.expanduser("~/.dAngr_history"))
        session = PromptSession(enable_history_search=True, history=cmd_history)
        if self.debug_file_path:
            dbg.handle(f"load {self.debug_file_path}", False)
        if self.script_path:
            try:
                if os.path.dirname(self.script_path):
                    os.chdir(os.path.dirname(self.script_path))
                proc:ScriptProcessor = ScriptProcessor(os.path.basename(self.script_path))
                first = True
                for line in proc.process_file():
                    if not line:
                        continue
                    #read script line by line and execute commands
                    with patch_stdout() as po:
                        if first:
                            prmpt2 = HTML(f'<darkcyan>(dAngr)> </darkcyan> {html.escape(line).strip()} <gray>(hit enter to proceed, Ctrl-c to end script)</gray>')
                            first = False
                        else:
                            l = line
                            if '\n' in line:
                                l = line.replace('\n', '\n' + " "*10)
                            prmpt2 = HTML(f'<darkcyan>(dAngr)> </darkcyan> {html.escape(l.strip())}')
                        try:
                            inp = session.prompt(prmpt2, completer=self.completer)
                            if not line.strip().startswith("#"):
                                # line = self._preprocess_input(conn, line)
                                if not dbg.handle(line, False):
                                    self.stop = True
                        except KeyboardInterrupt:
                            self.stop = True
                        except EOFError:
                            return # Ctrl-D to exit
                        except Exception as e:
                            if DEBUG_COMMANDS:
                                raise e
                            else:
                                conn.send_error(f"An unexpected error occurred: {e}")
                    if self.stop:
                        break
            except Exception as e:
                if DEBUG_COMMANDS:
                    raise e
                else:
                    conn.send_error(f"Error during script handling of {self.script_path}: {str(e)}")
                return
        self.stop = False
        self.script_path = None
        inp = ""

        while not self.stop:
            try:
                with patch_stdout() as po:
                    last_command = inp
                    inp = session.prompt(prmpt, completer=self.completer, validator=MyValidator(dbg))
                    if inp.strip() == "":
                        inp = last_command
                    lines = inp
                    if inp.strip() == "":
                        continue
                    if inp.rstrip().endswith(":"):
                        # process multiline input
                        lines = self.recusive_line_handler(session, lines, 4)

                    # lines = self._preprocess_input(conn, lines)
                    if not lines:
                        continue
                    if not dbg.handle(lines, False):
                        self.stop = True
                    self.reset_completer(dbg.context)
            except KeyboardInterrupt:
                continue
            except EOFError:
                return # Ctrl-D to exit
            except Exception as e:
                if DEBUG_COMMANDS:
                    raise e
                else:
                    conn.send_error(f"An unexpected error occurred: {e}")
    
    def recusive_line_handler(self, session: PromptSession, lines: str, line_indents: int) -> str:
        prompt_indents =  " "*(8 + line_indents)
        while True:
            line = session.prompt(prompt_indents + "", completer=self.completer)
            
            if line.strip() == "":
                return lines
            elif line.rstrip().endswith(":"):
                lines = self.recusive_line_handler(session, lines + f"\n{" "*line_indents}" + line, line_indents + 4)
            else:
                lines += f"\n{" "*line_indents}" + line

    
    def start_server(self):
        self.loop()




class BasicBlock:
    def __init__(self,address, size, instructions, assembly, function=None):
        self.address = address
        self.size = size
        self.instructions = instructions
        self.assembly = assembly
        self.function = function
    def __str__(self) -> str:
        #add function name if function is not None
        f = f"  Function: {self.function}\n" if self.function else ""
        return f"Address: {hex(self.address)}:\n{f}  Size: {self.size} bytes\n  Number of Instructions: {self.instructions}\n  Disassembly:\n    {str(self.assembly).replace('\n','\n    ')}"


class DebugSymbol:
    def __init__(self,name,type,address):
        self.name = name
        self.type = type
        self.address = address if address != 'N/A' else None
    def __str__(self) -> str:
        if self.address:
            return f"Debug Symbol: {self.name} ({self.type}) at {hex(self.address)}"
        else:
            return f"Debug Symbol: {self.name} ({self.type})"
        




from textwrap import indent
from typing import List

from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.cli.grammar.expressions import BREAK, BASECommand, Expression, Range, VariableRef
from dAngr.utils.utils import is_iterable
from .statements import Statement
from .script import Body

class ControlFlow(Statement):
    def __init__(self, body: Body):
        self.body: Body = body
    pass

class IfThenElse(ControlFlow):
    def __init__(self, condition: Expression, if_body: Body, else_body: Body|None = None):
        super().__init__(if_body)
        self.condition = condition
        self.else_body: Body = else_body if else_body else Body([])

    def __call__(self, context: ExecutionContext):
        if Expression.toBool(self.condition(context)):
            return self.body(ExecutionContext(context))
        elif self.else_body.statements:
            return self.else_body(ExecutionContext(context))
        return None

    def __repr__(self):
        return f"if {self.condition}:\n{indent(str(self.body), '   ')}\nelse:\n{indent(str(self.else_body), '   ')}"

    def __eq__(self, value: object) -> bool:
        return isinstance(value, IfThenElse) and self.condition == value.condition and self.body == value.body and self.else_body == value.else_body

class WhileLoop(ControlFlow):
    def __init__(self, condition: Expression, body: Body):
        super().__init__(body)
        self.condition = condition

    def __call__(self, context: ExecutionContext):
        while self.condition(context):
            r = self.body(ExecutionContext(context))
            if r == BREAK:
                break
            if isinstance(r, BASECommand) and r.base == "return":
                return r(context)

    def __repr__(self):
        return f"while {self.condition}:\n{indent(str(self.body), '   ')}"

    def __eq__(self, value: object) -> bool:
        return isinstance(value, WhileLoop) and self.condition == value.condition and self.body == value.body

class ForLoop(ControlFlow):
    def __init__(self, iterable: Expression, body: Body, item: VariableRef, index:VariableRef|None= None):
        super().__init__(body)
        self.index = index
        self.item = item
        self.iterable:Expression = iterable

    def __call__(self, context: ExecutionContext):
        if isinstance(self.iterable, Expression):
            iterable = self.iterable(context)
            #check if iterable is iterable
            if not is_iterable(iterable):
                raise ValueError(f"{iterable} is not iterable")
        else:
            iterable = self.iterable
        if self.index:
            for index, item in enumerate(iterable): # type: ignore
                ctx = ExecutionContext(context)
                ctx[self.index.name(context)] = index
                ctx[self.item.name(context)] = item
                r = self.body(ctx)
                if r == BREAK:
                    break
                if isinstance(r, BASECommand) and r.base == "return":
                    return r(context)
        else:
            for item in iterable: # type: ignore
                ctx = ExecutionContext(context)
                ctx[self.item.name(context)] = item
                r = self.body(ctx)
                if r == BREAK:
                    break
                if isinstance(r, BASECommand) and r.base == "return":
                    return r(context)

    def __repr__(self):
        if self.index:
            return f"for {self.index},{self.item} in {self.iterable}:\n{indent(str(self.body), '   ')}"
        else:
            return f"for {self.item} in {self.iterable}:\n{indent(str(self.body), '   ')}"

    def __eq__(self, value: object) -> bool:
        return isinstance(value, ForLoop) and self.index == value.index and self.item == value.item and self.iterable == value.iterable and self.body == value.body


# Generated from /workspaces/dAngr/src/dAngr/cli/grammar/dAngr.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .dAngrParser import dAngrParser
else:
    from dAngrParser import dAngrParser

# This class defines a complete generic visitor for a parse tree produced by dAngrParser.

class dAngrVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by dAngrParser#script.
    def visitScript(self, ctx:dAngrParser.ScriptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#statement.
    def visitStatement(self, ctx:dAngrParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#expression.
    def visitExpression(self, ctx:dAngrParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionRange.
    def visitExpressionRange(self, ctx:dAngrParser.ExpressionRangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionIn.
    def visitExpressionIn(self, ctx:dAngrParser.ExpressionInContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionNot.
    def visitExpressionNot(self, ctx:dAngrParser.ExpressionNotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionObject.
    def visitExpressionObject(self, ctx:dAngrParser.ExpressionObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionBool.
    def visitExpressionBool(self, ctx:dAngrParser.ExpressionBoolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionReference.
    def visitExpressionReference(self, ctx:dAngrParser.ExpressionReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionIf.
    def visitExpressionIf(self, ctx:dAngrParser.ExpressionIfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionAlt.
    def visitExpressionAlt(self, ctx:dAngrParser.ExpressionAltContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionParenthesis.
    def visitExpressionParenthesis(self, ctx:dAngrParser.ExpressionParenthesisContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ExpressionOperation.
    def visitExpressionOperation(self, ctx:dAngrParser.ExpressionOperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#assignment.
    def visitAssignment(self, ctx:dAngrParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#static_var.
    def visitStatic_var(self, ctx:dAngrParser.Static_varContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ext_command.
    def visitExt_command(self, ctx:dAngrParser.Ext_commandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#control_flow.
    def visitControl_flow(self, ctx:dAngrParser.Control_flowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#else_.
    def visitElse_(self, ctx:dAngrParser.Else_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#function_def.
    def visitFunction_def(self, ctx:dAngrParser.Function_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#body.
    def visitBody(self, ctx:dAngrParser.BodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#fstatement.
    def visitFstatement(self, ctx:dAngrParser.FstatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#iterable.
    def visitIterable(self, ctx:dAngrParser.IterableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#parameters.
    def visitParameters(self, ctx:dAngrParser.ParametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#condition.
    def visitCondition(self, ctx:dAngrParser.ConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#operation.
    def visitOperation(self, ctx:dAngrParser.OperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#py_basic_content.
    def visitPy_basic_content(self, ctx:dAngrParser.Py_basic_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#py_content.
    def visitPy_content(self, ctx:dAngrParser.Py_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#bash_content.
    def visitBash_content(self, ctx:dAngrParser.Bash_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#reference.
    def visitReference(self, ctx:dAngrParser.ReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#index.
    def visitIndex(self, ctx:dAngrParser.IndexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#identifier.
    def visitIdentifier(self, ctx:dAngrParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#numeric.
    def visitNumeric(self, ctx:dAngrParser.NumericContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#SlideStartLengthObject.
    def visitSlideStartLengthObject(self, ctx:dAngrParser.SlideStartLengthObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ReferenceObject.
    def visitReferenceObject(self, ctx:dAngrParser.ReferenceObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#BinaryStringObject.
    def visitBinaryStringObject(self, ctx:dAngrParser.BinaryStringObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#ListObject.
    def visitListObject(self, ctx:dAngrParser.ListObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#IndexedPropertyObject.
    def visitIndexedPropertyObject(self, ctx:dAngrParser.IndexedPropertyObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#DictionaryObject.
    def visitDictionaryObject(self, ctx:dAngrParser.DictionaryObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#NumericObject.
    def visitNumericObject(self, ctx:dAngrParser.NumericObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#SliceStartEndObject.
    def visitSliceStartEndObject(self, ctx:dAngrParser.SliceStartEndObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#StringObject.
    def visitStringObject(self, ctx:dAngrParser.StringObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#IDObject.
    def visitIDObject(self, ctx:dAngrParser.IDObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#PropertyObject.
    def visitPropertyObject(self, ctx:dAngrParser.PropertyObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#BoolObject.
    def visitBoolObject(self, ctx:dAngrParser.BoolObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#anything.
    def visitAnything(self, ctx:dAngrParser.AnythingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#anything_no.
    def visitAnything_no(self, ctx:dAngrParser.Anything_noContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#special_words.
    def visitSpecial_words(self, ctx:dAngrParser.Special_wordsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#range.
    def visitRange(self, ctx:dAngrParser.RangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#dangr_range.
    def visitDangr_range(self, ctx:dAngrParser.Dangr_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#bash_range.
    def visitBash_range(self, ctx:dAngrParser.Bash_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#python_range.
    def visitPython_range(self, ctx:dAngrParser.Python_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by dAngrParser#symbol.
    def visitSymbol(self, ctx:dAngrParser.SymbolContext):
        return self.visitChildren(ctx)



del dAngrParser



import asyncio
from unittest.mock import Mock, MagicMock, patch
from prompt_toolkit import PromptSession
import pytest

from dAngr.cli.command_line_debugger import DEBUGGER_COMMANDS, CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection
from dAngr.cli.server import Server


@pytest.fixture
def server():
    server = Server()
    server.start_server = Mock()
    # server.loop = Mock()
    return server


@patch('asyncio.run',new_callable=Mock)
def test_start_server(asyncio_run,server):

        def run_coroutine_sync(coro):
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(coro)

        asyncio.run(server.start_server())


        server.start_server()
        asyncio_run.assert_called_once()


@patch.object(CliConnection,'send_info')
@patch.object(CommandLineDebugger,'handle', side_effect=[False])
@patch.object(PromptSession ,'prompt', side_effect=["help"])
def test_loop(prompt, handle, send_info, server):
    
    server.loop()
    prompt.assert_called_once()
    send_info.assert_called_once_with("Welcome to dAngr, the symbolic debugger. Type help or ? to list commands.")
    handle.assert_called_once_with('help', False)


def test_check_shortnames(server):
    conn = CliConnection()
    dbg = CommandLineDebugger(conn)

            # get duplicate short commands and print short and fullname
    for cmd in DEBUGGER_COMMANDS.values():
        
        if cmd.name in ["start","continue"]:
            break
        assert not any([o.short_name and o.short_name == cmd.short_name for o in DEBUGGER_COMMANDS.values() if o != cmd]), f"Duplicate short command {cmd.short_name}"


def test_check_cmdnames(server):
    conn = CliConnection()
    dbg = CommandLineDebugger(conn)

            # get duplicate short commands and print short and fullname
    for cmd in DEBUGGER_COMMANDS.values():
        
        if cmd.name in ["start","continue"]:
            break
        assert not any([o.name and o.name == cmd.name for o in DEBUGGER_COMMANDS.values() if o != cmd]), f"Duplicate short command {cmd.name}"




import os
import re
import subprocess
from typing import Any, Callable, Dict, List, Tuple
import angr
from angr import SimCC, SimProcedure, SimulationManager, types
from angr.analyses.cfg.cfg_fast import CFGFast
from angr.knowledge_plugins.functions.function import Function
import angr.storage
import claripy

from dAngr.angr_ext.models import BasicBlock, DebugSymbol
from dAngr.angr_ext.step_handler import StepHandler, StopReason
from dAngr.cli.grammar.execution_context import Variable
from dAngr.cli.grammar.expressions import Constraint
from dAngr.cli.state_visualizer import StateVisualizer
from dAngr.utils.utils import AngrValueType, AngrObjectType, AngrType, DataType, DataType, Endness, SolverType, StreamType, SymBitVector, get_local_arch, remove_ansi_escape_codes
from dAngr.utils import utils
from .std_tracker import StdTracker
from .utils import create_entry_state, get_function_address, hook_simprocedures, load_module_from_file, get_function_by_name, get_function_by_addr
from .connection import Connection
from angr.storage.memory_mixins.convenient_mappings_mixin import *

from dAngr.exceptions import DebuggerCommandError
from dAngr.utils.loggers import get_logger
log = get_logger(__name__)
from dAngr.exceptions.InvalidArgumentError import InvalidArgumentError




class Debugger:
    def __init__(self, conn:Connection) -> None:
        self.conn = conn

        self._binary_path:str|None = None
        self._project:angr.project.Project|None = None
        self._simgr:angr.sim_manager.SimulationManager|None = None
        self._current_state:angr.SimState|None = None
        self._default_state_options = set()

        self._pause:bool = False

        self._base_addr:int = 0x400000
        self.verbose_step:bool = True
        self._function_prototypes = {}
        self._current_function:str = ''
        self._cfg:CFGFast|None = None
        self.stop_reason = StopReason.NONE
        self._save_unconstrained = False
        self._entry_point:int|Tuple[str,types.SimTypeFunction,SimCC,List[Any]]|None = None
        self._symbols:Dict[str,SymBitVector] = {}

    @property
    def project(self)->angr.project.Project:
        self.throw_if_not_initialized()
        return self._project # type: ignore
    @property
    def entry_point(self):
        return self._entry_point
    
    @property
    def cfg(self):
        if self._cfg is None:
            self.conn.send_info("Constructing cfg, this may take a while...")
            self._cfg = self.project.analyses.CFGFast(normalize=True)
        return self._cfg

    @property
    def simgr(self)->SimulationManager:
        if self._simgr is None:
            self.throw_if_not_initialized()
            if self._current_state is None:
                log.debug("Creating default simulation manager.")
                self.set_entry_state()
            self._simgr = self.project.factory.simulation_manager(self.current_state, save_unconstrained=self._save_unconstrained)
        return self._simgr
    
    @property
    def initialized(self):
        return self._project is not None
    @property
    def active(self,all=False):
        if all:
            return self._simgr and self._simgr.active and not self._simgr.deadended
        else: # there may be inactive states
            return self._simgr and self._simgr.active
    @property
    def finished(self, all=False):
        if all:
            return self._simgr and self._simgr.deadended and not self._simgr.active
        else: # there may be active states
            return self._simgr and self._simgr.deadended
    

    def unconstrained_fill(self, symbolic:bool ):
        if not symbolic:
            self._default_state_options = set([angr.options.ZERO_FILL_UNCONSTRAINED_MEMORY,angr.options.ZERO_FILL_UNCONSTRAINED_REGISTERS])            
        else:
            self._default_state_options = set([angr.options.SYMBOL_FILL_UNCONSTRAINED_MEMORY,angr.options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS])            

    @property
    def current_state(self):
        self.throw_if_not_initialized()
        if not self._current_state:
            if self.simgr.active:
                self._current_state = self.simgr.one_active          
            elif self.simgr.deadended:
                self._current_state = self.simgr.one_deadended
            else:
                raise DebuggerCommandError("No active or deadended states.")
        return self._current_state
    
    @current_state.setter
    def current_state(self, state:angr.SimState):
        self._set_current_state(state)

    def _set_current_state(self, state):
        if state:
            state.register_plugin('stdout_tracker', StdTracker())
        self._current_state = state
                
    def set_current_state(self, stateID:int, stash="active"):
        if stash not in self.simgr.stashes:
            raise DebuggerCommandError(f"Stash {stash} not found.")
        if stateID >= len(self.simgr.stashes[stash]):
            raise DebuggerCommandError(f"State with ID {stateID} not found in stash {stash}.")
        self._set_current_state(self.simgr.stashes[stash][stateID])

    def move_state_to_stash(self, stateID:int, from_stash:str, to_stash:str):
        if from_stash not in self.simgr.stashes:
            raise DebuggerCommandError(f"Stash {from_stash} not found.")
        if to_stash not in self.simgr.stashes:
            #create the stash
            self.simgr.stashes[to_stash] = []
        if stateID >= len(self.simgr.stashes[from_stash]):
            raise DebuggerCommandError(f"State with ID {stateID} not found in stash {from_stash}.")
        state = self.simgr.stashes[from_stash][stateID]
        self.simgr.move(from_stash, to_stash,lambda  s: s == state )
        return state
    
    def to_stash(self, stash:str):
        if not self.simgr.active:
            raise DebuggerCommandError("No active state to move.")
        self.simgr.move("active", stash, lambda s: s == self.current_state)

    def throw_if_not_initialized(self):
        if self._project is None:
            raise DebuggerCommandError("project not initialized.")
    def throw_if_not_active(self):
        if not self._current_state:
            raise DebuggerCommandError("no active states.")
    def get_source_info(self, addr):
        try:
            elf = self.project.loader.find_object_containing(addr) 
            src = max(elf.addr_to_line[addr]) # type: ignore
            return self.substitute_path_inverse(src[0]), src[1]
        except KeyError:
            return None,0
        
    def _find_elf_object_by_source(self, sourcefile, line_number):
        # Iterate through all ELF objects loaded by the project's loader
        for elf_obj in self.project.loader.all_elf_objects: 
            # Check if the ELF object has DWARF information
            if hasattr(elf_obj, 'addr_to_line'):
                # Attempt to reverse-map the source file and line number to an address
                for addr in elf_obj.addr_to_line:
                    # Use addr_to_line to get line information for each address
                    line_info = max(elf_obj.addr_to_line[addr])
                    # Check if the line information matches the source file and line number
                    if line_info[0] == sourcefile and line_info[1] == line_number:
                        # If a match is found, return the ELF object and the matching address
                        return elf_obj, addr
        # Return None if no matching ELF object or address is found
        return None
    def find_address(self, sourcefile, line_number)->int|None:
        n = self._find_elf_object_by_source(self.substitute_path(sourcefile), line_number)
        if n:
            return n[1]
        return None
    
    def set_call_state(self, func_addr:int, arguments):
        if self._simgr is not None:
            self.reset_state()
        # TODO: check add_options
        self._set_current_state( self.project.factory.call_state(func_addr,*arguments,
                    add_options = self._default_state_options, save_unconstrained=self._save_unconstrained))

    def set_entry_state(self,addr = None, *args, **kwargs):
        if self._simgr is not None:
            self.reset_state()
        if addr:
            kwargs['addr'] = addr
        if args:
            kwargs['args'] = args
        if kwargs.get('add_options') is None:
            kwargs['add_options'] = self._default_state_options
        self._set_current_state(self.project.factory.entry_state(
                     **kwargs))
    
    def set_full_state(self, *args, **kwargs):
        if self._simgr is not None:
            self.reset_state()
        if args:
            kwargs['args'] = args
        if kwargs.get('add_options') is None:
            kwargs['add_options'] = self._default_state_options
        self._set_current_state(self.project.factory.full_init_state(
                    **kwargs))

    def set_blank_state(self, *args, **kwargs):
        if self._simgr is not None:
            self.reset_state()
        if args:
            kwargs['args'] = args
        if kwargs.get('add_options') is None:
            kwargs['add_options'] = self._default_state_options
        log.debug(f"Creating blank state with kwargs: {kwargs}")
        state = self.project.factory.blank_state(**kwargs)
        self._set_current_state(state)
        
    @property
    def keep_unconstrained(self):
        return self._save_unconstrained
    
    @keep_unconstrained.setter  
    def keep_unconstrained(self, value:bool):
        self._save_unconstrained = value
        self._simgr = None

    def set_function_prototype(self, return_type:str, name:str, args:List[str]):  
        # Parse argument types
        sim_arg_types = [angr.types.parse_type(arg.strip()) for arg in args]
        sim_return_type = angr.types.parse_type(return_type)

        # Create function prototype
        self._function_prototypes[name] = {"prototype":angr.types.SimTypeFunction(sim_arg_types, sim_return_type)} # type: ignore
    def store_function(self, name, prototype, addr, cc):
        self._function_prototypes[name] = {"prototype":prototype, "addr":addr, "cc":cc}

     
    def get_stored_function(self, name:str):
        f = self._function_prototypes.get(name) if name in self._function_prototypes else None
        if not f:
            raise DebuggerCommandError(f"Function {name} not found.")
        return f
    
    def set_current_function(self, name:str):
        self._current_function = name
    
    def get_function_cc(self):
        cc = angr.default_cc(self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None)
        if cc is None:
            raise DebuggerCommandError("Failed to get calling convention.")
        return cc(self.project.arch)
    
    def get_function_info(self, func) -> Function |None:
       self.cfg
       if type(func) is int:
        return get_function_by_addr(self.project, func)
       else:
        return get_function_by_name(self.project, func)
    
    def get_function_prototype(self, prototype:str, arguments:List[str]):
        return angr.SimCC.guess_prototype(arguments, prototype).with_arch(self.project.arch)
    
    def get_function_address(self, name:str):
        return get_function_address(self.project, name)
    
    def get_function_callstate(self, function_name:str, prototype:types.SimTypeFunction, cc:SimCC,arguments:List[Any]):
        self._entry_point = (function_name,prototype,cc,arguments)
        if self._simgr is not None:
            self.reset_state()
        self._simgr,_ = create_entry_state(self.project,self._entry_point,self._default_state_options, self._current_state, self.veritesting)
        return self.simgr.one_active

    def get_variables(self, func_addr):
        names = list(self.project.kb.dvars._dvar_containers)
        pc = claripy.BVV(func_addr+self._base_addr,64) # self._simgr.active[0].ip
        vars = []
        for name in names:
            
            v = self.project.kb.dvars._dvar_containers[name].from_pc(pc)
            if v:
                vars.append(v)
        return vars


    def has_dwarf(self):
        try:
            return self.project.loader.main_object.has_dwarf_info
        except Exception as e:
            return False


    def substitute_path(self,path:str):
        path = os.path.abspath(os.path.normpath(path))
        if path.startswith(self.from_src_path):
            return self.to_src_path + path[len(self.from_src_path):]
        return path
        
    def substitute_path_inverse(self,path:str):
        path = os.path.abspath(os.path.normpath(path))
        if path.startswith(self.to_src_path):
            return self.from_src_path + path[len(self.to_src_path):]
        return path
    
    def init(self, binary_path:str, base_addr:int=0, from_src_path=None, to_src_path=None, veritesting:bool=False, **kwargs): #support passing custom angr arguments to project
        self.reset_state()
        self.veritesting = veritesting
        self._binary_path = binary_path
        self.from_src_path = os.path.abspath(os.path.expanduser(os.path.normpath(from_src_path))) if from_src_path else ''
        self.to_src_path = os.path.abspath(os.path.expanduser(os.path.normpath(to_src_path))) if to_src_path else ''
        if not os.path.exists(binary_path):
            raise InvalidArgumentError(f"File '{binary_path}' does not exist")
        
        main_opts = kwargs

        if base_addr:
            main_opts['base_addr'] = base_addr

        self._project = angr.Project(binary_path, load_options={'load_debug_info': True, 'auto_load_libs': False, 'main_opts':main_opts})
        if self.has_dwarf():
            self.project.kb.dvars.load_from_dwarf()
    
    def get_binary_info(self):
        # get general info such as path, name, arch, entry point, etc
        self.throw_if_not_initialized()
        if self._binary_path is None:
            raise DebuggerCommandError("Binary path not set.")
        
        return {
            "path":self._binary_path, 
            "arch":f" {self.project.arch.name}  ({str(self.project.arch.bits)} bits)", 
            "endian":self.project.arch.memory_endness.replace('Iend_', ''),
            "entry_point":hex(self.project.entry), 
            "base_addr": hex(self.project.loader.main_object.mapped_base)
        }
    
    def get_stdin_variables(self):
        return self.current_state.solver.get_variables('file', self.current_state.posix.stdin.ident)

    def get_binary_security_features(self):
        # get security features such as canary, nx, pie, etc
        self.throw_if_not_initialized()
        # run checksec as a process and parse the output
        features = {}
        try:
            checksec = subprocess.run(f"checksec --file={self._binary_path}", cwd=os.path.abspath(os.curdir), capture_output=True, text=True, shell=True)
            if checksec.returncode != 0:
                raise DebuggerCommandError(f"Failed to run checksec: {checksec.stdout}")
            output = checksec.stdout
            lines = output.split("\n")
            headers = re.split(r'\s{2,}|\t',lines[0])
            values = [x for x in (remove_ansi_escape_codes(r) for r in re.split(r'\s{2,}|\t',lines[1])) if x.strip()]
            for i in range(len(headers)):
                features[headers[i]] = values[i]

        except Exception as e:
            raise DebuggerCommandError(f"Failed to run checksec: {e}")
        return features
    
    def add_symbol(self, name:str, sym:SymBitVector):
        self._symbols[name] = sym

    def remove_symbol(self, name:str):
        if name in self._symbols:
            self._symbols.pop(name)
        else:
            raise DebuggerCommandError(f"Symbol {name} not found.")

    def get_symbol(self, name:str):
        if name in self._symbols:
            return self._symbols[name]
        else : raise DebuggerCommandError(f"Symbol {name} not found.")
    def to_symbol(self, name:str, lst:list):
        self.add_symbol(name, claripy.Concat(*lst))
        
    def find_symbol(self, name:str):
        if name in self._symbols:
            return self._symbols[name]
        return None
    
    def eval_symbol(self, sym:SymBitVector|Constraint, dtype:DataType, **kwargs):
        if isinstance(sym, SymBitVector):
            return self.current_state.solver.eval(sym, cast_to=dtype.to_type(), **kwargs)
        elif isinstance(sym, utils.Constraint):
            #drop endness from kwargs
            if 'endness' in kwargs:
                kwargs.pop('endness')
            return self.current_state.solver.eval(sym, cast_to=dtype.to_type(), **kwargs)
        else:
            raise DebuggerCommandError(f"Invalid symbol type {type(sym)}.")
        
    def eval_symbol_n(self, sym:SymBitVector|Constraint, n:int,solver_type:SolverType, dtype:DataType,  **kwargs):
        switch = {
            SolverType.UpTo: self.current_state.solver.eval_upto,
            SolverType.AtLeast: self.current_state.solver.eval_atleast,
            SolverType.Exact: self.current_state.solver.eval_exact
        }
        f = switch.get(solver_type)
        if not f:
            raise DebuggerCommandError(f"Invalid solver type {solver_type}.")
        if isinstance(sym, SymBitVector):
            return f(sym, n, cast_to=dtype.to_type(), **kwargs)
        elif isinstance(sym, utils.Constraint):
            #drop endness from kwargs
            if 'endness' in kwargs:
                kwargs.pop('endness')
            return f(sym, n, cast_to=dtype.to_type(), **kwargs)
        else:
            raise DebuggerCommandError(f"Invalid symbol type {type(sym)}.")

    def satisfiable(self, constraint=None):
        self.throw_if_not_active()
        return self.current_state.satisfiable(extra_constraints=[constraint])



    def set_symbol(self, name, value):
        self._symbols[name] = value
    def is_symbolic(self, value):
        self.throw_if_not_active()
        return self.current_state.solver.symbolic(value)

    def add_constraint(self, cs):
        self.throw_if_not_active()
        self.current_state.add_constraints(cs)
    def make_value(self, value:AngrObjectType)->AngrValueType:
        if isinstance(value, Variable):
            if not value.value:
                raise DebuggerCommandError("Variable has no value.")
            return value.value
        return value
    
    def add_to_stack(self, value:AngrObjectType):
        self.throw_if_not_active()
        v = self.make_value(value)
        self.current_state.stack_push(v)
        
    def get_stack(self, length, offset = 0):
        return self.current_state.stack_read(offset,length)
    
    def reset_state(self):
        self._simgr = None # state is reset upone requesting simgr
        self._current_state = None
        self.stop_reason = StopReason.NONE
        self._pause = False
        self._function_prototypes = {}
        self._current_function = ''
        self._cfg = None
        self.stop_reason = StopReason.NONE
        self._entry_point = None
#        self._symbols = {}

        log.info("State reset.")
    
    def stop(self):
        self._project = None
        self._simgr = None
        self._current_state = None
 
    # @param handler: StepHandler
    # @param check_until: callable return reason to stop, else None
    # @param exclude: callable returns True to exclude state from active stash
    
    def _run(self, handler:StepHandler, check_until:Callable[[angr.SimulationManager],StopReason] = lambda _:StopReason.NONE, exclude:Callable[[angr.SimState],bool] = lambda _:False, single:bool = False):
        self.throw_if_not_initialized()
        self.stop_reason:StopReason = StopReason.NONE
        #make sure current state is an active state and move it to the front if it is not
        if self.simgr.active:
            if self._current_state not in self.simgr.active:
                self._current_state = self.simgr.one_active
            else:
                #move current state to the front
                if self._current_state != self.simgr.one_active:
                    self.simgr.active.remove(self._current_state)
                    self.simgr.active.insert(0,self._current_state)
        else:
            raise DebuggerCommandError("No active states.")
        
        def selector_func(state):
            #return true if state is the first in the active stash
            return state == self.simgr.one_active  # type: ignore
        
        def filter_func(state):
            if state.addr == 0:
                return "deadended"
            #move excludes states to the exclude list
            if exclude and exclude(state):
                return "excluded"
            return None
        
        def step_func(simgr):
            # handle output
            if simgr.active:
                state = simgr.one_active
                std:StdTracker = state.get_plugin('stdout_tracker')
                std_data = std.get_new_string()
                if std_data:handler.handle_output(std_data)

        def until_func(simgr):
            if not simgr.active:
                self.stop_reason = StopReason.TERMINATE
                return True
            self.stop_reason = check_until(simgr)
            return self.stop_reason != StopReason.NONE
        self.simgr.run(stash="active", selector_func=selector_func, filter_func=filter_func,until=until_func,step_func=step_func, num_inst=1 if single else None)
        self._set_current_state( self.simgr.one_active if self.simgr.active else None)
        handler.handle_step(self.stop_reason, self._current_state)

    def back(self):
        #get the previous state
        if self.simgr.active:
            cur = self.simgr.one_active
            #remove the current state from the active stash
            self.simgr.active.remove(cur)
            #add parent state to the active stash
            if cur.history.parent:
                new_state = cur.history.parent.state.copy()
                self.simgr.active.insert(0,new_state)
                self._set_current_state(new_state)
            else:
                raise DebuggerCommandError("No parent state.")
        elif self.simgr.deadended:
            self._set_current_state(self.simgr.one_deadended)
  
    def get_current_addr(self):
        return self.current_state.addr


    def get_callstack(self,state):
        paths = [] 
        prev = state.addr
        i = 0
        for ix,s in enumerate(state.callstack):
            block = self.project.factory.block(prev)
            f = self.get_function_info(s.func_addr) if s.func_addr!=0 else None
            name = f.name if f else f"State at address{hex(s.func_addr)}"
            end = block.instruction_addrs[-1] if len(block.instruction_addrs) else s.func_addr # type: ignore
            paths.append({"addr":prev, "id":i, "func":s.func_addr, "end": end, "name":  name})
            prev = s.call_site_addr
            i += 1
        return paths

        
    def get_return_value(self)->int|None|claripy.ast.Base:
        prototype = self.get_stored_function(self._current_function)["prototype"]
        cc = self.get_stored_function(self._current_function)["cc"]
        if not cc or prototype.returnty is None:
            return None
        loc = cc.return_val(prototype.returnty)
        if loc is None:
            return None
        state = self.current_state
        val = loc.get_value(state, stack_base=state.regs.sp - cc.STACKARG_SP_DIFF)
        value = state.solver.eval(val, cast_to=int)
        return value
    
    def get_callable_function(self, addr:int):
        return self.project.factory.callable(addr)
 
    def get_bb_end_address(self, state):
        bb = state.block()
        if not bb.instruction_addrs:
            return state.addr
        return bb.instruction_addrs[-1]
    
    def get_bbs(self):
        # TODO: Add filtering support
        #return all basic blocks of the binary including block addr size instructions and capstone based on the CFG
        for node in self.cfg.graph.nodes():
            #include function name if available:
            func = self.cfg.kb.functions.get(node.function_address, None)
            yield BasicBlock(node.addr, node.size, len(node.instruction_addrs), node.block.capstone if node.block else None, func.name if func else None)
    def get_bb_count(self):
        return len(self.cfg.graph.nodes())
        
    def get_current_basic_block(self):
        state = self.current_state
        try:
            block = state.block()
            return BasicBlock(block.addr,block.size,block.instructions,block.capstone) # type: ignore
        except Exception as e:
            raise DebuggerCommandError(f"Failed to retrieve basic block: {e}")
        
    def get_stdstream(self, stream:StreamType):
        return self.current_state.posix.dumps(stream.value)
     
    def create_symbolic_file(self, name:str, content:str|claripy.ast.BV|claripy.ast.String|None=None, size:int|None=None):
        file = angr.storage.file.SimFile(name, content=content, size=size)
        self.current_state.fs.insert(name, file)
    
    def get_stashes(self):
        return list(self.simgr.stashes.keys())
            
    def list_paths(self, stash:str = "active"):
        if stash not in self.simgr.stashes:
            raise DebuggerCommandError(f"Stash {stash} not found.")
        return self.simgr.stashes[stash]
    
    def set_memory(self,address:int|SymBitVector,value:AngrValueType,size:int|None = None, endness:Endness = Endness.DEFAULT):
        if isinstance(value,str):
            # Encode string to bytes
            val = value.encode('utf-8')
        else:
            val = value
        en = Endness.to_arch_endness(endness, self.project)
        # Store the byte value in memory
        self.current_state.memory.store(address, val,size, endness=en,)
    
    def get_memory(self, address:int|SymBitVector, size:int|SymBitVector|None = None, endness:Endness = Endness.DEFAULT):
        state = self.current_state
        if address is int: 
            if size == 0:
                size = self.project.arch.bits // 8
        byte_value = state.memory.load(address, size, endness=Endness.to_arch_endness(endness, self.project))
        return byte_value
    
    def get_addr_for_symbol(self, symbol:str):
        # Letting the user know that certain options have to be set to allow this function.
        required_options = ["REVERSE_MEMORY_NAME_MAP", "TRACK_ACTION_HISTORY"]
        for option in required_options:
            if option not in self.current_state.options:
                raise DebuggerCommandError(f"Make sure you enable '{option}' when setting the entry state")

        return self.current_state.memory.addrs_for_name(symbol)
    
    def get_stream(self, stream:StreamType) -> str:
        state = self.current_state
        mapped = stream.name.lower()
        std:StdTracker = state.get_plugin(f'{mapped}_tracker')
        return std.get_prev_string()

    def cast_to(self, value:AngrValueType, dtype:DataType, **kwargs):
        return dtype.convert(value,  self.project.arch if self.initialized else get_local_arch(), **kwargs)    

    def load_hooks(self, filename):
        mod = load_module_from_file(filename)
        hooks = hook_simprocedures(self.project,mod)
        return hooks
    
    def add_hook(self, address:int, function:Callable, skip_length:int = 0, replace = True):
        self.project.hook(address, function, length=skip_length, replace=replace)

    def add_function_hook(self, target:int|str, function:SimProcedure):
        self.project.hook_symbol(target, function, replace=True)

    def add_to_state(self, name:str, value:AngrType):
        self.current_state.globals[name] = value
    def get_from_state(self, name:str):
        return self.current_state.globals[name]

    def get_constraints(self):
        return self.current_state.solver.constraints
    
    def get_binary_symbols(self):
        symbols = []
        for s in self.project.loader.symbols:
            symbols.append(DebugSymbol(s.name, s.type, s.rebased_addr))
        #remove duplicates
        symbols = list({symbol.name:symbol for symbol in symbols}.values())
        return symbols
    
    def get_decompiled_function(self, func_name):
        func = self.cfg.kb.functions.function(name=func_name)
        if func:
            decompiled = self.project.analyses.Decompiler(func)
            return decompiled.codegen.text if decompiled.codegen else None
        return None

    def get_decompiled_function_at_address(self, start, end):
        if not self._cfg:
            cfg = self.project.analyses.CFGFast(start=start,end=end, force_complete_scan=False, normalize=True)
        else:
            cfg = self._cfg
        decompiled = self.project.analyses.Decompiler(start)
        return decompiled.codegen.text if decompiled.codegen else None

    def get_binary_string_constants(self,min_length=4):
        # _ = self.cfg
        constants = []
        for section in self.project.loader.main_object.sections:
            if section.is_readable:  # Usually data sections
                base_addr = section.vaddr
                size = section.memsize
                if size == 0 or base_addr == 0:
                    continue
                data = self.project.loader.memory.load(base_addr, size)
                # Find all sequences of printable characters in the data
                # Adjust the regex pattern as needed for different types of strings
                re_pattern = rb'[ -~]{%d,}' % min_length
                matches = re.finditer(re_pattern, data)  # Find strings with at least 4 printable characters
                
                for match in matches:
                    # Compute the address of the string in the binary
                    start_offset = match.start()
                    end_offset = match.end()
                    string_value = match.group().decode('utf-8', errors='ignore')  # Decode to UTF-8
                    
                    # Add the string and its address
                    string_address = base_addr + start_offset
                    constants.append((hex(string_address), string_value))

        return constants
    def list_path_history(self, index:int = 0, stash="active"):
        # list basic blocks of states in the path history for both active and deadended states
        st = self.simgr.stashes[stash]
        if index >= len(st):
            raise DebuggerCommandError("Failed to find state")
        state = st[index]

        for a in state.history.recent_bbl_addrs:
                bb = self.project.factory.block(a)
                yield BasicBlock(bb.addr,bb.size,bb.instructions,bb.capstone)
    
    def _get_string_memory_from_state(self, address, state):
        string_array = []
        while True:
            # Read until a null byte is encountered
            endness = state.arch.memory_endness
            byte = state.memory.load(address, 1, endness=endness)
            if state.solver.is_true(byte == 0):
                break
            if not byte.concrete:
                log.error(f"Failed to get concrete value for byte at address {hex(address)}")
                break
            string_array.append(byte.concrete_value)
            address += 1
        str_value = ''.join([chr(byte) for byte in string_array])
        return str_value

    def get_string_memory(self, address):
        return self._get_string_memory_from_state(address, self.current_state)
        
    
    def list_registers(self):
        registers = self.project.arch.registers
        return registers
    
    def get_register_value(self, register):
        size = self.get_register(register)[1]
        value = self.current_state.registers.load(register, size)
        return value
    
    def get_register(self, register):
        if register not in self.project.arch.registers:
            raise DebuggerCommandError(f"Register '{register}' not found.")
        return self.current_state.registers.load(register)

    def set_register(self, register, value:int|claripy.ast.BV):
        self.current_state.registers.store(register, value)

    def visulize_state(self):
        state_printer = StateVisualizer(self.current_state)
        return state_printer.pprint()


from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection


import pytest


import os
from unittest.mock import Mock


class TestMemoryCommands:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)


    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        assert dbg.handle("load example")
        assert dbg.handle("unconstrained_fill False")
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("set_function_prototype 'int processMessage(char*, int, char**)'")
        assert dbg.handle("set_function_call 'processMessage(\"abc\",2,b\"0000000000\")'")
        conn.send_result = Mock()
        conn.send_info = Mock()
        conn.send_error = Mock()

        return dbg

    
    def test_get_memory(self, dbg, conn):
        assert dbg.handle("m = memory.get_memory 0x1000 3")
        assert dbg.handle("evaluate m")
        assert "b'abc'" == str(conn.send_result.call_args[0][0])
    
    # 
    # def test_get_int_memory(self, dbg, conn):
    #     assert dbg.handle("get_int_memory 0x3000 3")
    #     assert r == True
    #     assert "Memory at 0x2000: 0 (int)" == str(conn.send_info.call_args[0][0])
    
    def test_get_string_memory(self, dbg, conn):
        assert dbg.handle("get_memory_string 0x1000")
        assert "abc" == str(conn.send_result.call_args[0][0])
    
    def test_get_register(self, dbg, conn):
        assert dbg.handle("reg = get_register ip")
        assert dbg.handle("cast_to (evaluate reg) address")
        assert "0x40054d" == str(conn.send_result.call_args[0][0])

    
    def test_get_register_invalid(self, dbg, conn):
        assert dbg.handle("get_register invalid",False)
        assert "Register 'invalid' not found." == str(conn.send_error.call_args[0][0])
    
    
    def test_get_register_invalid_args(self, dbg, conn):
        assert dbg.handle("get_register",False)
        assert "missing a required argument: 'name'" == str(conn.send_error.call_args[0][0])
    
    
    def test_list_registers(self, dbg, conn):
        assert dbg.handle("list_registers")
        assert "Register rax " in str(conn.send_result.call_args[0][0])

    
    def test_set_register(self, dbg, conn):
        assert dbg.handle("set_register ip 0x4e")
        assert "Register ip: 0x4e." == str(conn.send_info.call_args[0][0])

    
    def test_set_memory_int(self, dbg, conn):
        assert dbg.handle("set_memory 0x1000 0x61")
        assert "Memory at 0x1000: 97." == str(conn.send_info.call_args[0][0])
        
    
    def test_set_memory_str(self, dbg, conn):
        assert dbg.handle("set_memory 0x1000 'abs'")
        assert "Memory at 0x1000: abs." == str(conn.send_info.call_args[0][0])

    
    def test_set_memory_bytes(self, dbg, conn):
        assert dbg.handle("set_memory 0x1000 b'1234'")
        assert "Memory at 0x1000: b'1234'." == str(conn.send_info.call_args[0][0])

    
    def test_unconstrained_fill_memory(self, dbg, conn):
        assert dbg.handle("unconstrained_fill")
        assert "Fill with zeros." == str(conn.send_info.call_args[0][0])

        assert dbg.handle("unconstrained_fill True")
        assert "Fill with symbols." == str(conn.send_info.call_args[0][0])
            
    
    def test_get_return_value(self, dbg, conn):
        assert dbg.handle("run")
        assert dbg.handle("run")
        assert dbg.handle("get_return_value")
        assert "6" == str(conn.send_result.call_args[0][0])



from claripy import List

from dAngr.cli.grammar.definitions import Definition, FunctionContext
from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.cli.grammar.expressions import BREAK, CONTINUE, BASECommand, VariableRef
from dAngr.cli.grammar.statements import Assignment, Statement

class StatementException(Exception):
    pass

class Body:
    def __init__(self, statements):
        from dAngr.cli.grammar.statements import Statement
        self.statements:List[Statement] = statements

    def is_static(self, statement):
        from dAngr.cli.grammar.statements import Assignment
        if isinstance(statement, Assignment) and isinstance(statement.variable, VariableRef):
            return statement.variable.is_static
        elif isinstance(statement, VariableRef):
            return statement.is_static
        return False

        
    def __call__(self, ctx:ExecutionContext):
        result = None
        if isinstance(ctx, FunctionContext):
            func_name = ctx.function.name
        else:
            func_name = "main"
        context = ExecutionContext(parent=ctx)
        try:
            statement_cnt = 0
            for s in self.statements:
                try:
                    if isinstance(s,BASECommand):
                        if s == BREAK:
                            return s
                        elif s == CONTINUE:
                            return
                        elif s.base == "return":
                            return s(context)
                    if self.is_static(s):
                        self._prepare_static(context, func_name, s)
                    result = s(context)
                    if result == BREAK:
                        return result
                    if isinstance(result, BASECommand) and result.base == "return":
                        return result(context)
                    statement_cnt += 1
                except StatementException as e:
                    raise e
                except Exception as e:
                    raise StatementException(f"Error executing statement {s} in function {func_name} line:{statement_cnt}\n\t{e}")
        finally:
            self._repair_static(context, func_name)
        return result
    
    def _prepare_static(self, context:ExecutionContext, func_name:str, s:Statement):
        var = None
        if isinstance(s, Assignment) and isinstance(s.variable, VariableRef):
            var = s.variable
        elif isinstance(s, VariableRef):
            var= s
        else:
            raise Exception("Unknown statement type")
        stat_name = var.static_name
        if context.root.find_variable(stat_name) is None:
            context.root.add_variable(stat_name, None)
        if isinstance(s, Assignment): #set global variable on each execution
            val = s.value(context)
            context.root[stat_name].value = val
        context.add_variable(var.name(context), context.root[stat_name].value)

    def _repair_static(self, context:ExecutionContext, func_name:str):
        #for all statements that are static, copy the value from context to the root context
        for s in self.statements:
            if self.is_static(s):
                var = None
                if isinstance(s, Assignment) and isinstance(s.variable, VariableRef):
                    var = s.variable
                elif isinstance(s, VariableRef):
                    var = s
                else:
                    raise Exception("Unknown statement type")
                context.root[var.static_name] = context[var.name].value

    def __repr__(self):
        return "\n".join([str(s) for s in self.statements])
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Body) and self.statements == value.statements

class Script(Body):
    def __init__(self, statements, definitions):
        super().__init__(statements)
        self.definitions:List[Definition] = definitions

    def __call__(self, context:ExecutionContext):
        for d in self.definitions:
            context.add_definition(d.name, d) # type: ignore
        result = None
        for s in self.statements:
            result = s(context)
        return result

    def __repr__(self):
        return f"Script({self.statements})"
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Script) and self.statements == value.statements and self.definitions == value.definitions



from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection

import pytest
import os
from unittest.mock import Mock


class TestBasicExecutionCommands:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)


    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("load 'example'")
        return dbg

    
    def test_reset(self, dbg, conn):
        assert dbg.handle("reset_state")
        conn.send_info.assert_called_with("State reset.")

    
    def test_continue(self, dbg, conn):
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")

    
    def test_load_hooks(self, dbg, conn):
        assert dbg.handle("load_hooks 'example_hooks.py'")
        conn.send_info.assert_called_with("Hooks 'example_hooks.py' successfully attached.")

    # pause needs further testing
    
    def test_pause(self, dbg, conn):
        assert dbg.handle("pause")
        conn.send_info.assert_called_with("Paused successfully.")

    
    def test_set_start_address(self, dbg, conn):
        assert dbg.handle("set_entry_state 0x40054d")
        conn.send_info.assert_called_with("Execution will start at address 0x40054d.")
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")
    
    
    def test_option_enum(self,dbg, conn):
        input = "set_entry_state add_options=[options.LAZY_SOLVES]"
        assert dbg.handle(input)
        conn.send_info.assert_called_with("Execution will start at specified entry point.")

    
    def test_start(self, dbg, conn):
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")

    
    def test_step_into(self, dbg, conn):
        assert dbg.handle("step")
        conn.send_info.assert_called_with("Stepped to: 0x4003f6.")

    
    def test_step_back(self, dbg, conn):
        assert dbg.handle("to_hex &state.addr")
        conn.send_result.assert_called_with("0x4003e0", True)
        assert dbg.handle("step")
        conn.send_info.assert_called_with("Stepped to: 0x4003f6.")
        assert dbg.handle("back")
        conn.send_info.assert_called_with("Stepped back to: 0x4003e0.")
        assert dbg.handle("step")
        conn.send_info.assert_called_with("Stepped to: 0x4003f6.")

    
    def test_step_out(self, dbg, conn):
        assert dbg.handle("step_out")
        conn.send_info.assert_called_with("Terminated.")


    
    def test_step_over(self, dbg, conn):
        assert dbg.handle("step_over")
        conn.send_info.assert_called_with("Terminated.")
    
    
    def test_select_path(self, dbg, conn):
        assert dbg.handle("select_state 0")
        assert 'Path 0 selected: 0x4003e0' in str(conn.send_info.call_args[0][0])
    
    
    def test_exclude_function(self, dbg, conn):
        dbg.handle("f = by_function 'main'")
        dbg.handle("exclude f")
        conn.send_info.assert_called_with("Function Filter: main added to exclusions.")
        assert dbg.handle("remove_exclusion_filter 0")
        conn.send_info.assert_called_with("Function Filter: main removed from exclusions.")
    
    
    def test_exclude_address(self, dbg, conn):
        assert dbg.handle("f = by_address 0x40054d")
        dbg.handle("exclude f")
        conn.send_info.assert_called_with("Address Filter: 0x40054d added to exclusions.")
        assert dbg.handle("f = by_address 0x40054d")
        dbg.handle("remove_exclusion_filter 0")
        conn.send_info.assert_called_with("Address Filter: 0x40054d removed from exclusions.")

    
    def test_list_exclusions(self, dbg, conn):
        assert dbg.handle("list_exclusions")
        conn.send_info.assert_called_with("No exclusions found.")
        dbg.handle("f = by_address 0x40054d")
        dbg.handle("exclude f")
        assert dbg.handle("list_exclusions")
        conn.send_result.assert_called_with("Exclusion(s): [0] Address Filter: 0x40054d", True)


import claripy
from dAngr.exceptions import DebuggerCommandError
from dAngr.utils.utils import Constraint, DataType, SolverType, SymBitVector, undefined, AngrType
from .base import BaseCommand

class SymbolCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)

    def add_symbol(self, name:str, size:int = 0, dtype:DataType = DataType.bytes):
        """
        Add a symbol of bytes with name and size.

        Args:
            name (str): Name of the symbol
            size (int): Size of the symbol
            dtype (DataType): type of the data, default is bytes.
        
        Short name: sa
        
        """
        if dtype == DataType.int:
            int_size = self.debugger.project.arch.sizeof["int"]
        else: int_size = size
        self.debugger.add_symbol(name, claripy.BVS(name, int_size*8))
        self.send_info(f"Symbol {name} created.")
        return self.debugger.get_symbol(name)
        
    def get_symbol (self, name:str):
        """
        Get a symbol by name.

        Args:
            name (str): Name of the symbol
        
        Short name: sym
        
        """
        return self.debugger.get_symbol(name)
    
    def chop_symbol(self, sym:str|SymBitVector, bits:int=1):
        """
        Chop a symbol to a specific number of bits.

        Args:
            sym (str|SymBitVector): Name of the symbol
            bits (int): Number of bits to chop
        
        Short name: cs
        
        """
        if isinstance(sym, str):
            sym = self.debugger.get_symbol(sym)
        return sym.chop(bits)
    
    def to_symbol(self, sym:str, list:list):
        """
        Concat items in list into a single symbol
        
        Args:
            sym (str): name of the symbol
            list (list): list of items
        """
        self.debugger.to_symbol(sym, list)
    
    def evaluate(self, sym:SymBitVector|str, dtype:DataType=DataType.bytes):
        """
        Evaluate a symbol and get the value.

        Args:
            sym (SymBitVector|str): Symbol to evaluate.
            dtype (DataType): Type of the symbol. Default is none.
        
        Short name: ev
        
        """
        if isinstance(sym, str):
            sym = self.debugger.get_symbol(sym)
        return self.debugger.eval_symbol(sym, dtype)
    
    def evaluate_n(self, sym:SymBitVector|str, n:int,  solve_info:SolverType=SolverType.UpTo,dtype:DataType=DataType.bytes):
        """
        Evaluate all symbols and get their values.

        Args:
            sym (SymBitVector|str): Symbol to evaluate.
            n (int): Number of symbols to evaluate.
            solve_info (SolverType): Type of the solver method. Default is UpTo. Possible Values: UpTo, AtLeast, Exact
            dtype (DataType): Type of the symbol. Default is none.
        
        Short name: eva
        
        """
        if isinstance(sym, str):
            sym = self.debugger.get_symbol(sym)
        return self.debugger.eval_symbol_n(sym, n, solve_info, dtype)
    
    def satisfiable(self, constraint:Constraint|None = None):
        """
        Check if the current state is satisfiable.

        Args:
            constraint (Constraint|None): Constraint to check. Default is None.
        
        Short name: sf
        
        """
        if constraint is None:
            return self.debugger.satisfiable()
        return self.debugger.satisfiable(constraint)
    def is_symbolic(self, sym:AngrType):
        """
        Check if a symbol is symbolic.

        Args:
            sym (AngrType): Name of the symbol
        
        Short name: sis
        
        """
        if isinstance(sym, str):
            sym = self.debugger.get_symbol(sym)
        if not isinstance(sym, SymBitVector):
            return False
        return self.debugger.is_symbolic(sym)
    # def add_symbol_int(self, name:str):
    #     """
    #     Add a symbol with name and size.

    #     Args:
    #         name (str): Name of the symbol
        
    #     Short name: sai
        
    #     """
    #     int_size = self.debugger.project.arch.sizeof["int"]
    #     self.debugger.add_symbol(name, claripy.BVS(name, int_size*8))
    #     self.send_info(f"Symbolic integer {name} created.")

    # def add_symbolic_string(self, name:str, size:int):
    #     """
    #     Add a symbol with name and size.

    #     Args:
    #         name (str): Name of the symbol
    #         size (int): Size of the symbol.
        
    #     Short name: sas
        
    #     """
    #     self.send_info(f"Symbolic string {name} created.")

    # def symbol_to_bytes(self, sym:str|SymBitVector):
    #     """
    #     Solve and get concrete symbol value in bytes based on current state.

    #     Args:
    #         sym (str|SymBitVector): Name of the symbol
        
    #     Short name: stb
        
    #     """
    #     if isinstance(sym, str):
    #         sym = self.debugger.get_symbol(sym)
    #     return self.debugger.cast_to(sym, DataType.bytes)
    
    # def symbol_to_int(self, sym:str|SymBitVector):
    #     """
    #     Solve and get concrete symbol value as int based on current state.

    #     Args:
    #         sym (str|SymBitVector): Name of the symbol
        
    #     Short name: sti
        
    #     """
    #     if isinstance(sym, str):
    #         sym = self.debugger.get_symbol(sym)
    #     return self.debugger.cast_to(sym, DataType.int)
    
    # def symbol_to_str(self, sym:str|SymBitVector):
    #     """
    #     Convert symbol to a str.

    #     Args:
    #         sym (str|SymBitVector): Name of the symbol
        
    #     Short name: sts
        
    #     """
    #     if isinstance(sym, str):
    #         sym = self.debugger.get_symbol(sym)
    #     return self.debugger.cast_to(sym, DataType.str)
    
    # def symbol_to_bool(self, sym:str|SymBitVector):
    #     """
    #     Solve and get concrete symbol value as bool based on current state.

    #     Args:
    #         sym (str|SymBitVector): Name of the symbol
        
    #     Short name: stB
        
    #     """
    #     if isinstance(sym, str):
    #         sym = self.debugger.get_symbol(sym)
    #     return self.debugger.cast_to(sym, DataType.bool)
    
    def remove_symbol(self, sym:str|SymBitVector):
        """
        Remove a symbol.

        Args:
            sym (str|SymBitVector): Name of the symbol
        
        Short name: sr
        
        """
        if isinstance(sym, SymBitVector):
            if len(sym.args) == 0:
                raise DebuggerCommandError("Symbol name not found.")
            name = sym.args[0]
        else: name = sym
        self.debugger.remove_symbol(name)
        self.send_info(f"Symbol {sym} removed.")
    
    # def set_symbol_value(self, sym:str|SymBitVector, value:int|bytes|str|SymBitVector):
    #     """
    #     Set a symbol value.

    #     Args:
    #         sym (str|SymBitVector): Name of the symbol
    #         value (int|bytes|str|SymBitVector): Value to set
        
    #     Short name: ssv
        
    #     """
    #     if isinstance(sym, str):
    #         sym = self.debugger.get_symbol(sym)
    #     value = self.to_value(value)
    #     self.debugger.set_symbol(sym, value)
    #     self.send_info(f"Symbol {sym} set to {value}.")

    def add_constraint(self, constraint:Constraint):
        """
        Add a constraint to a symbol.

        Args:
            constraint (Constraint): Constraint to add.
        
        Short name: sac
        
        """
        self.debugger.add_constraint(constraint)
        self.send_info(f"Constraint added.")
 